@@ -1,9 +1,10 @@
 /*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (c)  1985-2002, University of Amsterdam
+    E-mail:        jan@swi-prolog.org
+    WWW:           http://www.swi-prolog.org/packages/xpce/
+    Copyright (c)  1985-2025, University of Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -44,112 +45,71 @@
 
 					/* Win32 native locking */
 #ifdef USE_WIN32_CRITICAL_SECTION
-#define HAS_LOCK 1
+#define SYS_THREAD_T      DWORD
+#define SYS_LOCK_T        CRITICAL_SECTION
+#define SYS_THREAD_SELF() GetCurrentThreadId()
+#define SYS_LOCK(l)	  EnterCriticalSection(l)
+#define SYS_TRYLOCK(l)	  TryEnterCriticalSection(l)
+#define SYS_UNLOCK(l)	  LeaveCriticalSection(l)
 
-static CRITICAL_SECTION mutex;
-int lock_count;
-DWORD lock_owner;
+#define RECURSIVE_MUTEX_INIT { 0 }
 
-int
-pceMTTryLock(int lock)
-{ if ( XPCE_mt == TRUE )
-  { if ( TryEnterCriticalSection(&mutex) )	/* NT 4 and later! */
-    { if ( lock_count++ == 0 )
-	lock_owner = GetCurrentThreadId();
-      return TRUE;
-    } else
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-static inline void
-LOCK()
-{ if ( XPCE_mt == TRUE )
-  { EnterCriticalSection(&mutex);
-    if ( lock_count++ == 0 )
-      lock_owner = GetCurrentThreadId();
-  }
-}
-
-static inline void
-UNLOCK()
-{ if ( XPCE_mt == TRUE )
-  { if ( --lock_count == 0 )
-      lock_owner = 0;
-    LeaveCriticalSection(&mutex);
-  }
-}
-
-#define Code SWI_Code
-#include <SWI-Prolog.h>
-static foreign_t
-pce_lock_owner(term_t owner, term_t count)
-{ if ( PL_unify_integer(owner, lock_owner) &&
-       PL_unify_integer(count, lock_count) )
-    return TRUE;
-
-  return FALSE;
-}
-
-int
-pceMTinit()
-{ InitializeCriticalSection(&mutex);
-  XPCE_mt = TRUE;
-
-  PL_register_foreign("pce_lock_owner", 2, pce_lock_owner, 0);
-
-  succeed;
-}
-
-#endif /*USE_WIN32_CRITICAL_SECTION*/
-
-/* POSIX thread based locking */
-
-#if defined(_REENTRANT) && !defined(HAS_LOCK)
-#define HAS_LOCK 1
-#define var pthread_sys_var		/* avoid AIX name conflict */
+#else /*USE_WIN32_CRITICAL_SECTION*/
 #include <pthread.h>
-#undef var
 
-#ifdef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
-
-static pthread_mutex_t mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
-
-#define LOCK() \
-	if ( XPCE_mt ) pthread_mutex_lock(&mutex)
-#define UNLOCK() \
-	if ( XPCE_mt ) pthread_mutex_unlock(&mutex)
-
-int
-pceMTTryLock(int lock)
-{ if ( XPCE_mt == TRUE )
-  { if ( pthread_mutex_trylock(&mutex) != 0 )
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-#else /*PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP*/
-
-typedef struct _mutex_t
-{ pthread_t		owner;
-  int			count;
-  pthread_mutex_t	lock;
-} recursive_mutex_t;
+#define SYS_THREAD_T      pthread_t
+#define SYS_LOCK_T        pthread_mutex_t
+#define SYS_THREAD_SELF() pthread_self()
+#define SYS_LOCK(l)	  pthread_mutex_lock(l)
+#define SYS_TRYLOCK(l)	  (pthread_mutex_trylock(l) == 0)
+#define SYS_UNLOCK(l)	  pthread_mutex_unlock(l)
 
 #define RECURSIVE_MUTEX_INIT { 0, 0, PTHREAD_MUTEX_INITIALIZER }
 
+#endif /*USE_WIN32_CRITICAL_SECTION*/
+
+typedef struct _mutex_t
+{ SYS_THREAD_T	owner;
+  int		count;
+  SYS_LOCK_T	lock;
+} recursive_mutex_t;
+
 static recursive_mutex_t mutex = RECURSIVE_MUTEX_INIT;
+static bool redrawing = false;
+
+void
+setRedrawing(bool val)
+{ redrawing = val;
+}
+
+bool
+getRedrawing(void)
+{ return redrawing;
+}
+
+bool
+pceMTinit(void)
+{ XPCE_mt = TRUE;
+#ifdef USE_WIN32_CRITICAL_SECTION
+  InitializeCriticalSection(&mutex.lock);
+#endif
+  return true;
+}
+
+/* A Prolog thread has terminated */
+void
+pceMTdetach(void)
+{ destroyPceThreadData();
+}
 
 static inline void
-LOCK()
+LOCK(void)
 { if ( XPCE_mt )
-  { if ( mutex.owner != pthread_self() )
-    { pthread_mutex_lock(&(mutex.lock));
-      mutex.owner = pthread_self();
+  { SYS_THREAD_T self = SYS_THREAD_SELF();
+
+    if ( mutex.owner != self )
+    { SYS_LOCK(&(mutex.lock));
+      mutex.owner = self;
       mutex.count = 1;
     } else
       mutex.count++;
@@ -158,73 +118,87 @@ LOCK()
 
 
 static inline void
-UNLOCK()
+UNLOCK(void)
 { if ( XPCE_mt )
-  { if ( mutex.owner == pthread_self() )
-    { if ( --mutex.count < 1 )
+  { SYS_THREAD_T self = SYS_THREAD_SELF();
+
+    if ( mutex.owner == self )
+    { if ( --mutex.count == 0 )
       { mutex.owner = 0;
-	pthread_mutex_unlock(&(mutex.lock));
+	SYS_UNLOCK(&(mutex.lock));
       }
     } else
       assert(0);
   }
 }
 
+/* Public API */
+void
+pceMTLock(void)
+{ LOCK();
+}
 
-int
-pceMTTryLock(int lock)
+void
+pceMTUnlock(void)
+{ UNLOCK();
+}
+
+bool
+pceMTTryLock(void)
 { if ( XPCE_mt )
-  { if ( mutex.owner != pthread_self() )
-    { if ( pthread_mutex_trylock(&(mutex.lock)) != 0 )
-	return FALSE;
+  { SYS_THREAD_T self = SYS_THREAD_SELF();
 
-      mutex.owner = pthread_self();
+    if ( mutex.owner != self )
+    { if ( !SYS_TRYLOCK(&mutex.lock) )
+	return false;
+
+      mutex.owner = self;
       mutex.count = 1;
     } else
       mutex.count++;
   }
 
-  return TRUE;
+  return true;
 }
 
-#endif /*PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP*/
+
+/**
+ * Unlock/relock when we do a callback to Prolog.  This provides
+ * similar multi-threading as using the Python GIL.
+ */
 
 int
-pceMTinit()
-{ XPCE_mt = TRUE;
+pceMTUnlockAll(void)
+{ int rc = 0;
 
-  succeed;
-}
+  if ( XPCE_mt && !redrawing )
+  { SYS_THREAD_T self = SYS_THREAD_SELF();
 
-#endif /*_REENTRANT*/
+    if ( mutex.owner == self )
+    { assert(mutex.count);
+      rc = mutex.count;
+      mutex.owner = 0;
+      mutex.count = 0;
+      SYS_UNLOCK(&(mutex.lock));
+    }
+  }
 
-					/* No threading */
-#ifndef HAS_LOCK
-#define LOCK()
-#define UNLOCK()
-
-int					/* signal we can't do this */
-pceMTinit()
-{ fail;
-}
-
-int
-pceMTTryLock(int lock)
-{ return TRUE;
-}
-
-#endif
-
-void
-pceMTLock(int lock)
-{ LOCK();
+  return rc;
 }
 
 void
-pceMTUnlock(int lock)
-{ UNLOCK();
+pceMTRelock(int count)
+{ if ( XPCE_mt && count )
+  { SYS_THREAD_T self = SYS_THREAD_SELF();
+
+    assert(mutex.owner != self);
+    SYS_LOCK(&(mutex.lock));
+    mutex.owner = self;
+    mutex.count = count;
+  }
 }
 
+
 #define pushGoal(g) { LOCK(); \
 		      (g)->parent   = CurrentGoal; \
 		      CurrentGoal = g; \
@@ -673,17 +647,17 @@ status pceExecuteGoal(g)
 #define DEBUGGER(g) if ( PCEdebugging ) g
 #endif
 
-typedef Any (*ExecFunc0)(Any r, int vaac, Any *vaav);
-typedef Any (*ExecFunc1)(Any r, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc2)(Any r, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc3)(Any r, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc4)(Any r, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc5)(Any r, Any, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc6)(Any r, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc7)(Any r, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc8)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc9)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
-typedef Any (*ExecFunc10)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc0)(Any r, int vaac, Any *vaav);
+typedef Any (*VAGetFunc1)(Any r, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc2)(Any r, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc3)(Any r, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc4)(Any r, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc5)(Any r, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc6)(Any r, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc7)(Any r, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc8)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc9)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef Any (*VAGetFunc10)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
 
 typedef Any (*GetFunc0)(Any r);
 typedef Any (*GetFunc1)(Any r, Any);
@@ -698,6 +672,30 @@ typedef Any (*GetFunc9)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any);
 typedef Any (*GetFunc10)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any);
 typedef Any (*GetFunc11)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any);
 
+typedef status (*VASendFunc0)(Any r, int vaac, Any *vaav);
+typedef status (*VASendFunc1)(Any r, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc2)(Any r, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc3)(Any r, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc4)(Any r, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc5)(Any r, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc6)(Any r, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc7)(Any r, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc8)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc9)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+typedef status (*VASendFunc10)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any, int vaac, Any *vaav);
+
+typedef status (*SendFunc0)(Any r);
+typedef status (*SendFunc1)(Any r, Any);
+typedef status (*SendFunc2)(Any r, Any, Any);
+typedef status (*SendFunc3)(Any r, Any, Any, Any);
+typedef status (*SendFunc4)(Any r, Any, Any, Any, Any);
+typedef status (*SendFunc5)(Any r, Any, Any, Any, Any, Any);
+typedef status (*SendFunc6)(Any r, Any, Any, Any, Any, Any, Any);
+typedef status (*SendFunc7)(Any r, Any, Any, Any, Any, Any, Any, Any);
+typedef status (*SendFunc8)(Any r, Any, Any, Any, Any, Any, Any, Any, Any);
+typedef status (*SendFunc9)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any);
+typedef status (*SendFunc10)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any);
+typedef status (*SendFunc11)(Any r, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any, Any);
 
 status
 pceExecuteGoal(PceGoal g)
@@ -718,113 +716,203 @@ pceExecuteGoal(PceGoal g)
     else
       prof_node = NULL;
 
-    if ( m->function )
+    if ( m->function )				/* C-function implemented */
     { Any r      = g->receiver;
       Func f     = m->function;
       Any *a     = g->argv;
-      Any fval;
 
-      if ( g->va_type )
-      { int  vaac = g->va_argc;
-	Any *vaav = g->va_argv;
+      if ( g->flags & PCE_GF_GET )		/* Get method */
+      { Any fval;
 
-	switch(g->argc)
-	{ case 0:
-	    fval = (*(ExecFunc0)f)(r, vaac, vaav);
-	    break;
-	  case 1:
-	    fval = (*(ExecFunc1)f)(r, a[0], vaac, vaav);
-	    break;
-	  case 2:
-	    fval = (*(ExecFunc2)f)(r, a[0], a[1], vaac, vaav);
-	    break;
-	  case 3:
-	    fval = (*(ExecFunc3)f)(r, a[0], a[1], a[2], vaac, vaav);
-	    break;
-	  case 4:
-	    fval = (*(ExecFunc4)f)(r, a[0], a[1], a[2], a[3], vaac, vaav);
-	    break;
-	  case 5:
-	    fval = (*(ExecFunc5)f)(r, a[0], a[1], a[2], a[3], a[4], vaac, vaav);
-	    break;
-	  case 6:
-	    fval = (*(ExecFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], vaac, vaav);
-	    break;
-	  case 7:
-	    fval = (*(ExecFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
-				   a[6], vaac, vaav);
-	    break;
-	  case 8:
-	    fval = (*(ExecFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
-				   a[6], a[7], vaac, vaav);
-	    break;
-	  case 9:
-	    fval = (*(ExecFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
-				   a[6], a[7], a[8], vaac, vaav);
-	    break;
-	  case 10:
-	    fval = (*(ExecFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
-				    a[6], a[7], a[8], a[9], vaac, vaav);
-	    break;
-	  default:
-	    fval = (Any)FAIL;
-	    assert(0);
+	if ( g->va_type )
+	{ int  vaac = g->va_argc;
+	  Any *vaav = g->va_argv;
+
+	  switch(g->argc)
+	  { case 0:
+	      fval = (*(VAGetFunc0)f)(r, vaac, vaav);
+	      break;
+	    case 1:
+	      fval = (*(VAGetFunc1)f)(r, a[0], vaac, vaav);
+	      break;
+	    case 2:
+	      fval = (*(VAGetFunc2)f)(r, a[0], a[1], vaac, vaav);
+	      break;
+	    case 3:
+	      fval = (*(VAGetFunc3)f)(r, a[0], a[1], a[2], vaac, vaav);
+	      break;
+	    case 4:
+	      fval = (*(VAGetFunc4)f)(r, a[0], a[1], a[2], a[3], vaac, vaav);
+	      break;
+	    case 5:
+	      fval = (*(VAGetFunc5)f)(r, a[0], a[1], a[2], a[3], a[4], vaac, vaav);
+	      break;
+	    case 6:
+	      fval = (*(VAGetFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], vaac, vaav);
+	      break;
+	    case 7:
+	      fval = (*(VAGetFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], vaac, vaav);
+	      break;
+	    case 8:
+	      fval = (*(VAGetFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], a[7], vaac, vaav);
+	      break;
+	    case 9:
+	      fval = (*(VAGetFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], a[7], a[8], vaac, vaav);
+	      break;
+	    case 10:
+	      fval = (*(VAGetFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				      a[6], a[7], a[8], a[9], vaac, vaav);
+	      break;
+	    default:
+	      fval = (Any)FAIL;
+	      assert(0);
+	  }
+	} else
+	{ switch(g->argc)
+	  { case 0:
+	      fval = (*(GetFunc0)f)(r);
+	      break;
+	    case 1:
+	      fval = (*(GetFunc1)f)(r, a[0]);
+	      break;
+	    case 2:
+	      fval = (*(GetFunc2)f)(r, a[0], a[1]);
+	      break;
+	    case 3:
+	      fval = (*(GetFunc3)f)(r, a[0], a[1], a[2]);
+	      break;
+	    case 4:
+	      fval = (*(GetFunc4)f)(r, a[0], a[1], a[2], a[3]);
+	      break;
+	    case 5:
+	      fval = (*(GetFunc5)f)(r, a[0], a[1], a[2], a[3], a[4]);
+	      break;
+	    case 6:
+	      fval = (*(GetFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5]);
+	      break;
+	    case 7:
+	      fval = (*(GetFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6]);
+	      break;
+	    case 8:
+	      fval = (*(GetFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
+	      break;
+	    case 9:
+	      fval = (*(GetFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				    a[8]);
+	      break;
+	    case 10:
+	      fval = (*(GetFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				     a[8], a[9]);
+	      break;
+	    case 11:
+	      fval = (*(GetFunc11)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				     a[8], a[9], a[10]);
+	      break;
+	    default:
+	      fval = (Any)FAIL;
+	      assert(0);
+	  }
 	}
-      } else
-      { switch(g->argc)
-	{ case 0:
-	    fval = (*(GetFunc0)f)(r);
-	    break;
-	  case 1:
-	    fval = (*(GetFunc1)f)(r, a[0]);
-	    break;
-	  case 2:
-	    fval = (*(GetFunc2)f)(r, a[0], a[1]);
-	    break;
-	  case 3:
-	    fval = (*(GetFunc3)f)(r, a[0], a[1], a[2]);
-	    break;
-	  case 4:
-	    fval = (*(GetFunc4)f)(r, a[0], a[1], a[2], a[3]);
-	    break;
-	  case 5:
-	    fval = (*(GetFunc5)f)(r, a[0], a[1], a[2], a[3], a[4]);
-	    break;
-	  case 6:
-	    fval = (*(GetFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5]);
-	    break;
-	  case 7:
-	    fval = (*(GetFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6]);
-	    break;
-	  case 8:
-	    fval = (*(GetFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
-	    break;
-	  case 9:
-	    fval = (*(GetFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
-				  a[8]);
-	    break;
-	  case 10:
-	    fval = (*(GetFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
-				   a[8], a[9]);
-	    break;
-	  case 11:
-	    fval = (*(GetFunc11)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
-				   a[8], a[9], a[10]);
-	    break;
-	  default:
-	    fval = (Any)FAIL;
-	    assert(0);
+	g->rval = fval;
+	rval = !!fval;
+      } else					/* Send method */
+      {	if ( g->va_type )
+	{ int  vaac = g->va_argc;
+	  Any *vaav = g->va_argv;
+
+	  switch(g->argc)
+	  { case 0:
+	      rval = (*(VASendFunc0)f)(r, vaac, vaav);
+	      break;
+	    case 1:
+	      rval = (*(VASendFunc1)f)(r, a[0], vaac, vaav);
+	      break;
+	    case 2:
+	      rval = (*(VASendFunc2)f)(r, a[0], a[1], vaac, vaav);
+	      break;
+	    case 3:
+	      rval = (*(VASendFunc3)f)(r, a[0], a[1], a[2], vaac, vaav);
+	      break;
+	    case 4:
+	      rval = (*(VASendFunc4)f)(r, a[0], a[1], a[2], a[3], vaac, vaav);
+	      break;
+	    case 5:
+	      rval = (*(VASendFunc5)f)(r, a[0], a[1], a[2], a[3], a[4], vaac, vaav);
+	      break;
+	    case 6:
+	      rval = (*(VASendFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], vaac, vaav);
+	      break;
+	    case 7:
+	      rval = (*(VASendFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], vaac, vaav);
+	      break;
+	    case 8:
+	      rval = (*(VASendFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], a[7], vaac, vaav);
+	      break;
+	    case 9:
+	      rval = (*(VASendFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				     a[6], a[7], a[8], vaac, vaav);
+	      break;
+	    case 10:
+	      rval = (*(VASendFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5],
+				      a[6], a[7], a[8], a[9], vaac, vaav);
+	      break;
+	    default:
+	      rval = FAIL;
+	      assert(0);
+	  }
+	} else
+	{ switch(g->argc)
+	  { case 0:
+	      rval = (*(SendFunc0)f)(r);
+	      break;
+	    case 1:
+	      rval = (*(SendFunc1)f)(r, a[0]);
+	      break;
+	    case 2:
+	      rval = (*(SendFunc2)f)(r, a[0], a[1]);
+	      break;
+	    case 3:
+	      rval = (*(SendFunc3)f)(r, a[0], a[1], a[2]);
+	      break;
+	    case 4:
+	      rval = (*(SendFunc4)f)(r, a[0], a[1], a[2], a[3]);
+	      break;
+	    case 5:
+	      rval = (*(SendFunc5)f)(r, a[0], a[1], a[2], a[3], a[4]);
+	      break;
+	    case 6:
+	      rval = (*(SendFunc6)f)(r, a[0], a[1], a[2], a[3], a[4], a[5]);
+	      break;
+	    case 7:
+	      rval = (*(SendFunc7)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6]);
+	      break;
+	    case 8:
+	      rval = (*(SendFunc8)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
+	      break;
+	    case 9:
+	      rval = (*(SendFunc9)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				    a[8]);
+	      break;
+	    case 10:
+	      rval = (*(SendFunc10)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				     a[8], a[9]);
+	      break;
+	    case 11:
+	      rval = (*(SendFunc11)f)(r, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7],
+				     a[8], a[9], a[10]);
+	      break;
+	    default:
+	      rval = FAIL;
+	      assert(0);
+	  }
 	}
       }
-
-					/* end of function-implemation */
-
-      if ( fval )
-      { if ( g->flags & PCE_GF_GET )
-	  g->rval = fval;
-	rval = SUCCEED;
-      } else
-	rval = FAIL;
+					/* end of function-implementation */
     } else				/* not a C-function */
     { if ( objectIsInstanceOf(m->message, ClassCode) )
       {					/* A function object */
@@ -1444,5 +1532,3 @@ resolveGetMethodObject(Any obj, Class class, Name sel, Any *receiver)
 
   fail;
 }
-
-
